package org.bdware.sc.boundry.utils;

import org.bdware.sc.ContractProcess;
import org.bdware.sc.compiler.PermissionStub;
import org.bdware.sc.node.Permission;
import org.rocksdb.Options;
import org.rocksdb.RocksDB;
import org.rocksdb.RocksDBException;
import org.rocksdb.RocksIterator;
import wrp.jdk.nashorn.internal.runtime.PropertyMap;
import wrp.jdk.nashorn.internal.runtime.ScriptObject;
import wrp.jdk.nashorn.internal.scripts.JO;

import java.io.File;

@PermissionStub(permission = Permission.RocksDB)
public class RocksDBUtil {
    static {
        RocksDB.loadLibrary();
    }

    RocksDB rocksDB;

    public RocksDBUtil(String path, boolean readOnly) {
        try {
            Options options = new Options();
            options.setCreateIfMissing(true);
            File parent = new File("./ContractDB/" + ContractProcess.instance.getContractName());
            path = new File(parent, path).getAbsolutePath();
            File lockFile = new File(path, "LOCK");
            lockFile.delete();
            if (readOnly) {
                rocksDB = RocksDB.openReadOnly(options, path);
            } else rocksDB = RocksDB.open(options, path);
        } catch (RocksDBException e) {
            e.printStackTrace();
        }
    }

    public static RocksDBUtil loadDB(String path, boolean readOnly) {
        return new RocksDBUtil(path, readOnly);
    }

    public static RocksDBUtil loadDB(String path, String readOnly) {
        return new RocksDBUtil(path, Boolean.parseBoolean(readOnly));
    }

    public String get(String key) {
        try {
            return new String(rocksDB.get(key.getBytes()));
        } catch (Exception e) {
            e.printStackTrace();
        }
        return null;
    }

    public long estimateKeySize() {
        try {
            return rocksDB.getLongProperty("rocksdb.estimate-num-keys");
        } catch (Exception e) {
            e.printStackTrace();
            return 0;
        }
    }

    public void put(String key, String value) {
        try {
            rocksDB.put(key.getBytes(), value.getBytes());
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public String delete(String key) {
        try {
            rocksDB.delete(key.getBytes());

            return "success";
        } catch (Exception e) {
            e.printStackTrace();
        }
        return "failed";
    }

    public RocksIterator newIterator() {
        try {
            return rocksDB.newIterator();
        } catch (Exception e) {
            e.printStackTrace();
        }
        return null;
    }

    public ScriptObject getNext(RocksIterator iter) {

        if (iter.isValid()) {
            JO ret = new JO(PropertyMap.newMap());
            ret.put("key", new String(iter.key()), false);
            ret.put("value", new String(iter.value()), false);
            iter.next();
            return ret;
        }
        return null;
    }
}
